/*
 * Icon generation helpers for slide puzzle game
 * Copyright © 2022 Nick Bowler
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <config.h>
#include <string.h>

#include "colour.h"
#include "icon.h"

static void format_row(unsigned char *row, unsigned long c,
                       unsigned s, unsigned y, unsigned w, unsigned h)
{
	unsigned x;

	for (x = 0; x < w; x++) {
		row[x] = (c >> 8*COLOUR_PRIMARY) & 0xff;
		if (x < s || y < s)
			row[x] = (c >> 8*COLOUR_LIGHT) & 0xff;
		if ((x+s >= w && x+y >= w) || (y+s >= h && x+y >= h))
			row[x] = (c >> 8*COLOUR_DARK) & 0xff;
	}
}

/*
 * The 16x16 icon is drawn with 1px shadow, 6x6 tiles, with 1 pixel cropped
 * off all the edge tiles.
 */
void icon_tile16(unsigned char *icon, unsigned long c, int tile_x, int tile_y)
{
	int out_x, out_y, out_w, y;
	unsigned char row[6];

	out_x = tile_x * 11 / 2;
	out_y = tile_y * 11 / 2;
	out_w = (5 + (tile_x == 1)) * sizeof row[0];

	for (y = 0+(tile_y == 0); y < 6-(tile_y == 2); y++) {
		format_row(row, c, 1, y, 6, 6);
		memcpy(&icon[16 * out_y++ + out_x], &row[tile_x == 0], out_w);
	}
}

/*
 * The 24x24 icon is drawn with 1px shadow and 8x8 tiles.
 */
void icon_tile24(unsigned char *icon, unsigned long c, int tile_x, int tile_y)
{
	int out_x, out_y, y;

	out_x = tile_x * 8;
	out_y = tile_y * 8;

	for (y = 0; y < 8; y++) {
		format_row(&icon[24 * out_y++ + out_x], c, 1, y, 8, 8);
	}
}

/*
 * The 32x32 icon is drawn with 1px shadow with slightly uneven tiles on
 * an 11-10-11 pixel grid.
 */
void icon_tile32(unsigned char *icon, unsigned long c, int tile_x, int tile_y)
{
	int out_x, out_y, out_w, out_h, y;

	out_x = 10*tile_x + (tile_x > 0);
	out_y = 10*tile_y + (tile_y > 0);
	out_w = 10 + (tile_x != 1);
	out_h = 10 + (tile_y != 1);

	for (y = 0; y < out_h; y++) {
		format_row(&icon[32 * out_y++ + out_x], c, 1, y, out_w, out_h);
	}
}

/*
 * The 48x48 icon is drawn with 2px shadow and 16x16 tiles.
 */
void icon_tile48(unsigned char *icon, unsigned long c, int tile_x, int tile_y)
{
	int out_x, out_y, y;

	out_x = tile_x * 16;
	out_y = tile_y * 16;

	for (y = 0; y < 16; y++) {
		format_row(&icon[48 * out_y++ + out_x], c, 2, y, 16, 16);
	}
}
